from dataclasses import dataclass

import abc
import enum
import typing as t
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from kupala.json import jsonify
from kupala.requests import Request

SCOPE_KEY = 'flash_messages'
SESSION_KEY = 'flash_messages'


class MessageCategory(enum.Enum):
    DEBUG = 'debug'
    INFO = 'info'
    SUCCESS = 'success'
    WARNING = 'warning'
    ERROR = 'error'


@dataclass
class FlashMessage:
    category: str
    message: str

    def __str__(self) -> str:
        return self.message


class FlashBag:
    Category = MessageCategory

    def __init__(self, messages: list[FlashMessage] = None):
        self._messages: list[FlashMessage] = messages or []

    def add(self, message: str, category: t.Union[MessageCategory, str]) -> None:
        if isinstance(category, MessageCategory):
            category = category.value
        self._messages.append(FlashMessage(category, message))

    def get_by_category(self, category: t.Union[MessageCategory, str]) -> list[FlashMessage]:
        if isinstance(category, MessageCategory):
            category = category.value

        messages = [message for message in self._messages if message.category == category]
        self._messages = [message for message in self._messages if message.category != category]
        return messages

    def debug(self, message: str) -> None:
        self.add(message, MessageCategory.DEBUG)

    def info(self, message: str) -> None:
        self.add(message, MessageCategory.INFO)

    def success(self, message: str) -> None:
        self.add(message, MessageCategory.SUCCESS)

    def warning(self, message: str) -> None:
        self.add(message, MessageCategory.WARNING)

    def error(self, message: str) -> None:
        self.add(message, MessageCategory.ERROR)

    def all(self) -> t.List[FlashMessage]:
        return self._messages

    def consume(self) -> list[FlashMessage]:
        """Return all messages and empty the bag."""
        messages = self._messages.copy()
        self._messages.clear()
        return messages

    def clear(self) -> None:
        self._messages = []

    def __len__(self) -> int:
        return len(self._messages)

    def __iter__(self) -> t.Iterator[FlashMessage]:
        return iter(self.consume())

    def __bool__(self) -> bool:
        return len(self) > 0


class MessageStorage(abc.ABC):  # pragma: nocover
    @abc.abstractmethod
    def load(self, scope: Scope) -> FlashBag:
        raise NotImplementedError()

    @abc.abstractmethod
    def save(self, scope: Scope, bag: FlashBag) -> None:
        raise NotImplementedError()


class SessionStorage(MessageStorage):
    def load(self, scope: Scope) -> FlashBag:
        if 'session' not in scope:
            raise KeyError('Sessions are disabled. Flash messages depend on SessionMiddleware.')

        return FlashBag(
            [FlashMessage(message['category'], message['message']) for message in scope['session'].get(SESSION_KEY, [])]
        )

    def save(self, scope: Scope, bag: FlashBag) -> None:
        scope['session'][SESSION_KEY] = jsonify(bag.all())


_storage_map: t.Dict[str, t.Type[MessageStorage]] = {
    'session': SessionStorage,
}


class FlashMessagesMiddleware:
    def __init__(self, app: ASGIApp, storage: t.Union[MessageStorage, t.Literal["session"]] = 'session') -> None:
        self.app = app
        if isinstance(storage, str):
            storage_class = _storage_map[storage]
            storage = storage_class()
        self.storage = storage

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope['type'] != 'http':  # pragma: nocover
            await self.app(scope, receive, send)
            return

        bag = self.storage.load(scope)

        async def send_wrapper(message: Message) -> None:
            if message['type'] == 'http.response.start':
                self.storage.save(scope, bag)
            await send(message)

        scope[SCOPE_KEY] = bag
        scope['state'].get('template_context', {}).update({'messages': bag})

        await self.app(scope, receive, send_wrapper)


def flash(request: Request) -> FlashBag:
    assert SCOPE_KEY in request.scope, 'Flash messages require FlashMessagesMiddleware.'
    return request.scope[SCOPE_KEY]



from __future__ import annotations

import importlib
import inspect
import os.path
import re
import typing as t
from starlette.concurrency import run_in_threadpool

CAMEL_TO_SNAKE_PATTERN = re.compile(r'(?<!^)(?=[A-Z])')


def camel_to_snake(name: str) -> str:
    return CAMEL_TO_SNAKE_PATTERN.sub('_', name).lower()


def import_string(path: str, package: str = None) -> t.Any:
    attr = None
    if ':' in path:
        module_name, attr = path.split(':')
    else:
        module_name = path

    module_instance = importlib.import_module(module_name, package)
    if attr:
        return getattr(module_instance, attr)
    return module_instance


def resolve_path(path: t.Union[str, os.PathLike]) -> str:
    path = str(path)
    if not path.startswith('@'):
        return os.path.abspath(path)
    package_name, _, package_path = path.replace('@', '').partition(os.sep)
    package_spec = importlib.import_module(package_name)
    return os.path.join(str(os.path.dirname(str(package_spec.__file__))), package_path)


async def run_async(fn: t.Callable, *args: t.Any, **kwargs: t.Any) -> t.Any:
    if inspect.iscoroutinefunction(fn):
        return await fn(*args, **kwargs)
    return await run_in_threadpool(fn, *args, **kwargs)



import asyncio
import hashlib
import html
import inspect
import os
import sys
import traceback
import typing
from pprint import pformat
from starlette.concurrency import run_in_threadpool
from starlette.requests import Request
from starlette.responses import HTMLResponse, PlainTextResponse, Response
from starlette.types import ASGIApp, Message, Receive, Scope, Send

STYLES = """
:root {
    --red: rgb(239 68 68);
    --yellow-lightest: rgb(254 249 195);
    --gray: rgb(17 24 39);
    --gray-light: rgb(107 114 128);
    --gray-lighter: rgb(229 231 235);
    --gray-lightest: rgb(243 244 246);
}

html,
body {
    font-size: 14px;
    line-height: 1.6;
    background: #fff;
    padding: 0;
    margin: 0;
    color: var(--gray);
    background: var(--gray-lightest);
    font-family: Inter var, ui-sans-serif, system-ui,
        -apple-system, BlinkMacSystemFont, Segoe UI, Roboto,
        Helvetica Neue, Arial, Noto Sans, sans-serif, Apple Color Emoji,
        Segoe UI Emoji, Segoe UI Symbol, Noto Color Emoji;
}

.text-muted {
    color: var(--gray-light);
}

.bg-white {
    background-color: white;
}

pre {
    font-family: consolas, monospace;
    margin: 0;
}

.header {
    padding: 48px 48px 0 48px;
    color: var(--gray-light);
}

.exception-class {
    color: var(--red);
}

.exception {
    font-size: 1.5rem;
    color: var(--red);
}

section {
    padding: 24px 48px;
}

section.trace {
    padding-top: 48px;
    padding-bottom: 48px;
    display: grid;
    grid-template-columns: repeat(2, minmax(0, 1fr));
    gap: 24px;
}

.snippet-wrapper {
    display: none;
}

.snippet-wrapper.current {
    display: block;
}

.snippet {
    background-color: white;
    padding: 0 16px;
}

.snippet header {
    padding: 16px 0;
    border-bottom: 1px solid var(--gray-lightest);
}

.snippet footer {
    padding: 16px 0;
    border-top: 1px solid var(--gray-lightest);
    display: flex;
    align-items: center;
    justify-content: space-between;
    color: var(--gray-light);
}

.snippet pre {
    margin: 2px 0;
    white-space: pre-wrap;
    overflow-x: auto;
    overflow-wrap: anywhere;
}

.snippet .line {
    display: flex;
    overflow-x: auto;
    overflow-wrap: normal;
}

.snippet .line.highlight {
    background-color: var(--yellow-lightest);
}

.snippet .line-number {
    text-align: right;
    width: 48px;
    margin-right: 24px;
    color: var(--gray-light);
    flex-shrink: 0;
}

.snippet .code {
    flex: auto;
}

.frames .frame {
    display: flex;
    align-items: center;
    background-color: transparent;
    cursor: pointer;
    padding: 4px 8px;
    border-radius: 4px;
    gap: 12px;
    margin-bottom: 4px;
}

.frames .frame:hover {
    background-color: var(--gray-lighter);
}

.frames .frame.current {
    background-color: white;
}

.dot {
    height: 10px;
    width: 10px;
    border-radius: 100%;
    background-color: var(--gray-light);
}

.dot-red {
    background-color: var(--red);
}

details {
    margin-bottom: 16px;
}

summary {
    cursor: pointer;
}

dl {
    display: grid;
    grid-template-columns: repeat(12, minmax(0, 1fr));
    align-items: center;
    margin: 0;
    padding: 4px;
    border-top: 1px solid var(--gray-lighter);
}

dl:first-of-type {
    margin-top: 16px;
}

dl:hover {
    background-color: var(--gray-lightest);
}

dt {
    color: var(--gray-light);
    text-overflow: ellipsis;
    grid-column: span 2 / span 2;
}

dd {
    flex: auto;
    grid-column: span 10 / span 10;
    overflow-x: auto;
}

.locals {
    padding: 2px;
}

.locals:hover {
    background-color: transparent;
}

.locals dt {
    padding: 2px;
}

.locals dd {
    flex: auto;
    overflow-x: auto;
    padding: 2px 8px;
    background-color: var(--gray-lightest);
}

.open-uri {
    font-weight: bold;
    color: var(--gray);
    text-decoration: none;
}

@media(max-width: 1200px) {
    section {
        padding: 12px 16px;
    }

    section.trace {
        grid-template-columns: repeat(1, minmax(0, 1fr));
    }

    dl {
        grid-template-columns: repeat(1, minmax(0, 1fr));
    }

    dd {
        margin-inline-start: 0px;
    }

    .package-dir {
        display: none;
    }
}

"""

JS = """
<script type="text/javascript">
    Array.from(document.querySelectorAll('[id^="switch"]')).forEach(function(el) {
        el.addEventListener('click', function() {
            var current = document.querySelector('.snippet-wrapper.current');
            if (current) {
                current.classList.remove('current');
            }

            var frameIndex = el.dataset.frameIndex;
            var target = document.querySelector('#snippet-' + frameIndex);
            if (target) {
                target.classList.add('current');
            }

            var currentSwitch = document.querySelector('.frame.current');
            if (currentSwitch) {
                currentSwitch.classList.remove('current');
            }
            el.classList.add('current')
        });
    });
</script>
"""

TEMPLATE = """
<!DOCTYPE html>
<html lang="en">
    <head>
        <title>{exception}: {error} | Starlette Debugger</title>
        <meta name="viewport" content="width=device-width">
        <style type='text/css'>
            {styles}
        </style>
    </head>
    <body>
        {header}
        <main>
            <section class="trace">
                {code_snippet}
                <div class="frames">
                {frames}
                </div>
            </section>
            <section class="bg-white">
                {request}
                {headers}
                {session}
                {cookies}
                {request_state}
                {app_state}
                {platform}
                {environment}
            </section>
        </main>
        {js}
    </body>
</html>
"""

HEADER_TEMPLATE = """
<header class="header">
    <div>
        <b class="exception-class">{exception_class}</b> at {method} {path}
    </div>
    <div class="exception text-red">{message}</div>
</header>
"""

LOCALS_ROW_TEMPLATE = """
<dl class="locals">
    <dt>{name}</dt>
    <dd><pre>{value}</pre></dd>
</dl>
"""

CODE_LINE_TEMPLATE = """
<pre class="line {highlight}">
    <div class="line-number">{line_number}</div>
    <div class="code">{line}</div>
</pre>
"""

CODE_TEMPLATE = """
<div id="snippet-{id}" class="snippet-wrapper{extra_css_class}">
    <div class="snippet">
        <header>
            <a class="open-uri" href="{open_path}"><span class="package-dir">{package_dir}</span>{relative_path}</b></a>
        </header>
        <div>
            {lines}
        </div>
        {locals}
        <footer>
            <div class="symbol">{package}.{symbol}</div>
            <div class="package">{package}</div>
        </footer>
    </div>
</div>
"""

FRAME_LINE_TEMPLATE = """
<div class="frame {extra_css_class}" id="switch-{id}" data-frame-index="{id}">
    <div class="dot{dot}"></div>
    <div>
        <span class="text-muted">{package}</span>
        <span title="{file_path}">{file_name}<span>
        <span class="text-muted">in</span>
        <span class="symbol">{symbol}</span>
        <span class="text-muted">at line</span> {line_number}
    </div>
</div>
"""

DETAILS_ROW_TEMPLATE = """
<dl>
    <dt>{label}</dt>
    <dd>{value}</dd>
</dl>
"""

DETAILS_TEMPLATE = """
<details {open}>
    <summary>{label}</summary>
    {rows}
</details>
"""


def get_relative_filename(path: str) -> str:
    for sys_path in reversed(sorted(sys.path)):
        if sys_path in path:
            path = path.replace(sys_path + "/", "")
            break
    return path


def format_qual_name(obj: typing.Any) -> str:
    if inspect.isclass(obj):
        module_name = obj.__module__
        type_name = obj.__name__
    else:
        module_name = obj.__class__.__module__
        type_name = obj.__class__.__name__

    if module_name == "builtins":
        module_name = ""
    if module_name:
        return f"{module_name}.{type_name}"
    return type_name


def frame_id(frame: inspect.FrameInfo) -> str:
    hash = hashlib.md5()
    hash.update(f"{frame.filename}{frame.lineno}".encode())
    return hash.hexdigest()


def is_vendor(frame: inspect.FrameInfo) -> bool:
    return sys.exec_prefix in frame.filename


def get_package_name(frame: inspect.FrameInfo) -> str:
    return frame.frame.f_globals.get("__package__", '').split(".")[0]


def get_symbol(frame: inspect.FrameInfo) -> str:
    symbol = ""
    if "self" in frame.frame.f_locals:
        symbol = frame.frame.f_locals["self"].__class__.__name__

    if "cls" in frame.frame.f_locals:
        symbol = frame.frame.f_locals["cls"].__name__

    # if we cannot detect class name then just a method name will be rendered
    function = frame.function
    if symbol:
        function = symbol + "." + function
    return function


def mask_secrets(key: str, value: str) -> str:
    key = key.lower()
    if any(
        [
            "key" in key,
            "token" in key,
            "password" in key,
            "secret" in key,
        ]
    ):
        return "*" * 8
    return value


EDITORS = {'pycharm': 'pycharm://open?file={path}&line={line}'}


def generate_file_uri(path: str, line: int, editor: str) -> str:
    uri = EDITORS.get(editor, '').format(path=path, line=line)
    if not uri:
        uri = 'file://' + path
    return uri


class ServerErrorMiddleware:
    """
    Handles returning 500 responses when a server error occurs.

    If 'debug' is set, then traceback responses will be returned,
    otherwise the designated 'handler' will be called.

    This middleware class should generally be used to wrap *everything*
    else up, so that unhandled exceptions anywhere in the stack
    always result in an appropriate 500 response.
    """

    def __init__(
        self,
        app: ASGIApp,
        handler: typing.Callable = None,
        debug: bool = False,
        editor: typing.Optional[str] = os.environ.get('EDITOR', ''),
    ) -> None:
        self.app = app
        self.handler = handler
        self.debug = debug
        self.editor = editor

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        response_started = False

        async def _send(message: Message) -> None:
            nonlocal response_started, send

            if message["type"] == "http.response.start":
                response_started = True
            await send(message)

        try:
            await self.app(scope, receive, _send)
        except Exception as exc:
            if not response_started:
                request = Request(scope)
                if self.debug:
                    # In debug mode, return traceback responses.
                    response = self.debug_response(request, exc)
                elif self.handler is None:
                    # Use our default 500 error handler.
                    response = self.error_response(request, exc)
                else:
                    # Use an installed 500 error handler.
                    if asyncio.iscoroutinefunction(self.handler):
                        response = await self.handler(request, exc)
                    else:
                        response = await run_in_threadpool(self.handler, request, exc)

                await response(scope, receive, send)

            # We always continue to raise the exception.
            # This allows servers to log the error, or allows test clients
            # to optionally raise the error within the test case.
            raise exc

    def format_line(self, index: int, line: str, frame_lineno: int, frame_index: int) -> str:
        return CODE_LINE_TEMPLATE.format(
            line=html.escape(line).replace(" ", "&nbsp"),
            line_number=(frame_lineno - frame_index) + index,
            highlight=" highlight" if index == frame_index else "",
        )

    def generate_frame_html(self, frame: inspect.FrameInfo, is_collapsed: bool) -> str:
        symbol = get_symbol(frame)
        extra_css_classes = []
        if not is_collapsed:
            extra_css_classes.append("current")
        if is_vendor(frame):
            extra_css_classes.append("vendor")

        return FRAME_LINE_TEMPLATE.format(
            id=frame_id(frame),
            dot="" if is_vendor(frame) else " dot-red",
            package="{package} &middot;".format(package=get_package_name(frame)) if is_vendor(frame) else "",
            file_path=html.escape(frame.filename),
            file_name=html.escape(get_relative_filename(frame.filename)),
            symbol=symbol,
            line_number=frame.lineno,
            extra_css_class=" ".join(extra_css_classes),
        )

    def generate_locals_for_frame(self, frame: inspect.FrameInfo) -> str:
        return "".join(
            LOCALS_ROW_TEMPLATE.format(name=var_name, value=html.escape(pformat(var_value, indent=2)))
            for var_name, var_value in frame.frame.f_locals.items()
        )

    def generate_snippet_for_frame(self, frame: inspect.FrameInfo, is_collapsed: bool) -> str:
        code_context = "".join(
            self.format_line(index, line, frame.lineno, frame.index)  # type: ignore
            for index, line in enumerate(frame.code_context or [])
        )

        return CODE_TEMPLATE.format(
            id=frame_id(frame),
            package_dir=frame.filename.replace(get_relative_filename(frame.filename), ""),
            relative_path=html.escape(get_relative_filename(frame.filename)),
            file_path=html.escape(frame.filename),
            lines=code_context,
            extra_css_class="" if is_collapsed else " current",
            locals=self.generate_locals_for_frame(frame),
            symbol=get_symbol(frame),
            package=get_package_name(frame),
            open_path=generate_file_uri(frame.filename, frame.lineno, self.editor or ''),
        )

    def render_details_row(self, data: typing.Mapping) -> str:
        row_html = ""
        for name, value in data.items():
            row_html += DETAILS_ROW_TEMPLATE.format(label=name, value=mask_secrets(name, value))

        if not row_html:
            row_html = "<dl><dt>empty</dt><dd></dd></dl>"
        return row_html

    def generate_request_details_html(self, request: Request) -> str:
        return DETAILS_TEMPLATE.format(
            label="Request",
            rows=self.render_details_row(
                {
                    "Method": request.method,
                    "Path": request.url.path,
                    "Path params": "<br>".join(
                        [f'<span class="text-muted">{k}</span> = {v}' for k, v in request.path_params.items()]
                    ),
                    "Query params": "<br>".join(
                        [f'<span class="text-muted">{k}</span> = {v}' for k, v in request.query_params.items()]
                    ),
                    "Content type": request.headers.get("Content-Type", ""),
                    "Client": f"{request.client.host}:{request.client.port}",
                },
            ),
            open="open",
        )

    def generate_headers_html(self, request: Request) -> str:
        return DETAILS_TEMPLATE.format(
            label="Headers",
            rows=self.render_details_row(request.headers),
            open="",
        )

    def generate_cookies_html(self, request: Request) -> str:
        return DETAILS_TEMPLATE.format(
            label="Cookies",
            rows=self.render_details_row(request.cookies),
            open="",
        )

    def generate_session_html(self, request: Request) -> str:
        if "session" in request.scope:
            return DETAILS_TEMPLATE.format(
                label="Session",
                rows=self.render_details_row(request.headers),
                open="",
            )
        return ""

    def generate_environment_html(self, request: Request) -> str:
        return DETAILS_TEMPLATE.format(
            label="Environment",
            rows=self.render_details_row(os.environ),
            open="",
        )

    def generate_platform_html(self, request: Request) -> str:
        return DETAILS_TEMPLATE.format(
            label="Platform",
            rows=self.render_details_row(
                {
                    "Python version": sys.version,
                    "Platform": sys.platform,
                    "Python": sys.executable,
                    "Paths": "<br>".join(sys.path),
                }
            ),
            open="",
        )

    def generate_request_state_html(self, request: Request) -> str:
        return DETAILS_TEMPLATE.format(
            label="Request state",
            rows=self.render_details_row({k: v for k, v in request.state._state.items()}),
            open="",
        )

    def generate_app_state_html(self, request: Request) -> str:
        if "app" in request.scope:
            return DETAILS_TEMPLATE.format(
                label="App state",
                rows=self.render_details_row({k: v for k, v in request.app.state._state.items()}),
                open="",
            )
        return ""

    def generate_html(self, request: Request, exc: Exception, limit: int = 7) -> str:
        traceback_obj = traceback.TracebackException.from_exception(exc, capture_locals=True)

        header_html = HEADER_TEMPLATE.format(
            exception_class=format_qual_name(traceback_obj.exc_type),
            method=request.method,
            message=html.escape(str(exc) or '""'),
            path=request.url.path,
        )

        frames_html = ""
        code_html = ""
        is_collapsed = False
        exc_traceback = exc.__traceback__
        if exc_traceback is not None:
            frames = inspect.getinnerframes(exc_traceback, limit)
            for frame in reversed(frames):
                frames_html += self.generate_frame_html(frame, is_collapsed)
                code_html += self.generate_snippet_for_frame(frame, is_collapsed)
                is_collapsed = True

        return TEMPLATE.format(
            styles=STYLES,
            js=JS,
            header=header_html,
            code_snippet=code_html,
            frames=frames_html,
            exception=traceback_obj.exc_type.__name__,
            error=html.escape(str(traceback_obj)),
            request_state=self.generate_request_state_html(request),
            app_state=self.generate_app_state_html(request),
            platform=self.generate_platform_html(request),
            request=self.generate_request_details_html(request),
            headers=self.generate_headers_html(request),
            session=self.generate_session_html(request),
            cookies=self.generate_cookies_html(request),
            environment=self.generate_environment_html(request),
        )

    def generate_plain_text(self, exc: Exception) -> str:
        return "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))

    def debug_response(self, request: Request, exc: Exception) -> Response:
        accept = request.headers.get("accept", "")

        if "text/html" in accept:
            content = self.generate_html(request, exc)
            return HTMLResponse(content, status_code=500)
        content = self.generate_plain_text(exc)
        return PlainTextResponse(content, status_code=500)

    def error_response(self, request: Request, exc: Exception) -> Response:
        return PlainTextResponse("Internal Server Error", status_code=500)



import os
import pathlib
import typing as t
from deesk.drivers.fs import LocalFsDriver
from deesk.storage import Storage as BaseStorage


class Storage(BaseStorage):  # pragma: nocover
    async def url(self, path: t.Union[str, os.PathLike]) -> str:
        raise NotImplementedError()

    def abspath(self, path: t.Union[str, os.PathLike]) -> str:
        raise NotImplementedError()


class LocalStorage(Storage):
    def __init__(self, base_dir: t.Union[str, os.PathLike]) -> None:
        self.base_dir = base_dir
        super().__init__(driver=LocalFsDriver(base_dir=base_dir))

    async def url(self, path: t.Union[str, os.PathLike]) -> str:
        return str(path)

    def abspath(self, path: t.Union[str, os.PathLike]) -> str:
        return str(pathlib.Path(self.base_dir) / path)


class S3Storage(Storage):
    def __init__(self, link_ttl: int = 300, **kwargs: t.Any) -> None:
        from deesk.drivers.s3 import S3Driver

        self.link_ttl = link_ttl
        self.driver = S3Driver(**kwargs)
        super().__init__(self.driver)

    async def url(self, path: t.Union[str, os.PathLike]) -> str:
        async with self.driver.session.client('s3', endpoint_url=self.driver.endpoint_url) as client:
            return await client.generate_presigned_url(
                'get_object',
                Params={'Bucket': self.driver.bucket, 'Key': path},
                ExpiresIn=self.link_ttl,
            )

    def abspath(self, path: t.Union[str, os.PathLike]) -> str:
        return ''



from __future__ import annotations

import os
import typing
from imia import BaseAuthenticator, LoginManager, UserProvider
from mailers import Encrypter, Mailer, Plugin, Signer, create_transport_from_url

from kupala.contracts import PasswordHasher, TemplateRenderer
from kupala.storages.storages import LocalStorage, S3Storage, Storage
from kupala.utils import import_string

if typing.TYPE_CHECKING:
    from kupala.app.base import BaseApp


class Extension:
    def initialize(self, app: BaseApp) -> None:
        pass


_PasswordHasherType = typing.Literal['pbkdf2_sha256', 'pbkdf2_sha512', 'argon2', 'bcrypt', 'des_crypt'] | PasswordHasher


class Passwords(Extension):
    def __init__(self, backend: _PasswordHasherType = 'pbkdf2_sha256') -> None:
        self._manager = self._create(backend)

    def hash(self, plain_password: str) -> str:
        return self._manager.hash(plain_password)

    def verify(self, plain: str, hashed: str) -> bool:
        return self._manager.verify(plain, hashed)

    def use(self, backend: _PasswordHasherType) -> None:
        self._manager = self._create(backend)

    def _create(self, backend: _PasswordHasherType) -> PasswordHasher:
        if isinstance(backend, str):
            imports = {
                'pbkdf2_sha256': 'passlib.handlers.pbkdf2:pbkdf2_sha256',
                'pbkdf2_sha512': 'passlib.handlers.pbkdf2:pbkdf2_sha512',
                'argon2': 'passlib.handlers.argon2:argon2',
                'bcrypt': 'passlib.handlers.bcrypt:bcrypt',
                'des_crypt': 'passlib.handlers.des_crypt:des_crypt',
            }
            return import_string(imports[backend])
        return backend


class Renderer(Extension):
    def __init__(self, renderer: TemplateRenderer | None = None) -> None:
        self._template_renderer = renderer

    def use(self, renderer: TemplateRenderer) -> None:
        self._template_renderer = renderer

    def render(self, template_name: str, context: dict[str, typing.Any] = None) -> str:
        assert self._template_renderer, 'Template rendering is not configured.'
        return self._template_renderer.render(template_name, context)


class Mail(Extension):
    def __init__(self) -> None:
        self._mailers: dict[str, Mailer] = {}

    def use(
        self,
        url: str,
        *,
        from_address: str = 'no-reply@example.com',
        from_name: str = 'Example',
        signer: Signer = None,
        encrypter: Encrypter,
        plugins: list[Plugin] = None,
        name: str = 'default',
    ) -> None:
        if from_name:
            from_address = f'{from_name} <{from_address}>'
        transport = create_transport_from_url(url)
        self.add(
            name,
            Mailer(
                transport,
                from_address=from_address,
                plugins=plugins,
                signer=signer,
                encrypter=encrypter,
            ),
        )

    def add(self, name: str, mailer: Mailer) -> None:
        self._mailers[name] = mailer


class Authentication(Extension):
    def __init__(self, app: BaseApp) -> None:
        self.user_provider: UserProvider | None = None
        self.authenticators: list[BaseAuthenticator] = []
        self.password_verifier = app.passwords
        self.secret_key = app.secret_key

    @property
    def login_manager(self) -> LoginManager:
        assert self.user_provider, 'Authentication component requires user provider to be configured.'
        return LoginManager(
            user_provider=self.user_provider,
            password_verifier=self.password_verifier,
            secret_key=self.secret_key,
        )


class Storages(Extension):
    def __init__(self, storages: dict[str, Storage] = None) -> None:
        self.default = 'default'
        self._storages: dict[str, Storage] = storages or {}

    def get(self, name: str) -> Storage:
        return self._storages[name]

    def get_default(self) -> Storage:
        if self.default in self._storages:
            return self._storages[self.default]
        if len(self._storages) == 1:
            return list(self._storages.values())[0]
        raise KeyError('No default storage configured.')

    def add(self, name: str, storage: Storage) -> None:
        self._storages[name] = storage

    def add_local(self, name: str, base_dir: str | os.PathLike) -> None:
        self.add(name, LocalStorage(base_dir))

    def add_s3(
        self,
        name: str,
        bucket: str,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        region_name: str = None,
        profile_name: str = None,
        endpoint_url: str = None,
        link_ttl: int = 300,
    ) -> None:
        self.add(
            name,
            S3Storage(
                bucket=bucket,
                aws_secret_access_key=aws_secret_access_key,
                aws_access_key_id=aws_access_key_id,
                region_name=region_name,
                profile_name=profile_name,
                endpoint_url=endpoint_url,
                link_ttl=link_ttl,
            ),
        )



from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.exceptions import ValidationError
from kupala.middleware.exception import default_validation_error_handler
from kupala.requests import Request


class FormErrorsMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert 'session' in scope, 'FormErrorsMiddleware requires sessions enabled.'
        assert 'flash_messages' in scope, 'FormErrorsMiddleware requires flash messaging enabled.'

        try:
            await self.app(scope, receive, send)
        except ValidationError as exc:
            request = Request(scope, receive, send)
            response = await default_validation_error_handler(request, exc)
            await response(scope, receive, send)



import typing
from starlette import status
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from kupala.responses import PlainTextResponse


class LargeEntityError(ValueError):
    pass


class RequestLimitMiddleware:
    """Limit request body to a value of max_body_size.
    When request body exceeds the limit then 413 response returned.
    """

    def __init__(self, app: ASGIApp, max_body_size: typing.Optional[int] = 2048) -> None:
        self.app = app
        self.max_body_size = max_body_size

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        total_len = 0
        if scope['type'] != 'http':
            return await self.app(scope, receive, send)

        async def receive_wrapper() -> Message:
            message = await receive()
            if message['type'] != 'http.request' or self.max_body_size is None:
                return message

            nonlocal total_len
            body = message.get('body', b'')
            total_len += len(body)
            if total_len > self.max_body_size:
                raise LargeEntityError()
            return message

        try:
            await self.app(scope, receive_wrapper, send)
        except LargeEntityError:
            response = PlainTextResponse('Entity Too Large', status_code=status.HTTP_413_REQUEST_ENTITY_TOO_LARGE)
            await response(scope, receive, send)



from __future__ import annotations

import abc
import typing

import imia
from imia.authentication import APIKeyAuthenticator, BaseAuthenticator, BearerAuthenticator, HTTPBasicAuthenticator, \
    SessionAuthenticator, TokenAuthenticator
from imia.impersonation import exit_impersonation, get_original_user, impersonate, impersonation_is_active
from imia.login import get_session_auth_hash, get_session_auth_id, login_user
from imia.protocols import Authenticator, UserLike
from imia.user_providers import InMemoryProvider, UserProvider
from imia.user_token import AnonymousUser, LoginState

from kupala.requests import Request

__all__ = [
    'LoginManager', 'UserLike', 'Authenticator', 'UserToken', 'InMemoryProvider', 'UserProvider',
    'BaseAuthenticator', 'BaseUser', 'BearerAuthenticator', 'TokenAuthenticator', 'SessionAuthenticator',
    'HTTPBasicAuthenticator', 'APIKeyAuthenticator', 'impersonate', 'impersonation_is_active', 'exit_impersonation',
    'get_original_user', 'get_session_auth_id', 'get_session_auth_hash', 'login_user', 'AnonymousUser', 'LoginState',
]


class LoginManager(imia.LoginManager):
    @classmethod
    def from_request(cls, request: Request) -> LoginManager:
        return request.app.auth.login_manager


class UserToken(imia.UserToken):
    @classmethod
    def from_request(cls, request: Request) -> UserToken:
        return request.auth


class BaseUser(abc.ABC):  # pragma: nocover
    @abc.abstractmethod
    def get_id(self) -> typing.Any:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_display_name(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_scopes(self) -> list[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_hashed_password(self) -> str:
        raise NotImplementedError()



import inspect
import jinja2
import typing as t
from starlette.concurrency import run_in_threadpool
from starlette.exceptions import HTTPException as BaseHTTPException
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from kupala.exceptions import HTTPException, ValidationError
from kupala.requests import Request
from kupala.responses import GoBackResponse, HTMLResponse, JSONResponse, PlainTextResponse, Response

E = t.TypeVar('E', bound=Exception)
ErrorHandler = t.Callable[[Request, E], t.Any]
_renderer = jinja2.Environment(loader=jinja2.PackageLoader(__name__.split('.')[0]))


async def default_validation_error_handler(request: Request, exc: ValidationError) -> Response:
    if request.wants_json:
        return JSONResponse({'message': exc.message, 'errors': exc.errors}, 400)

    await request.remember_form_data()
    request.set_form_errors(dict(exc.errors or {}))
    response = GoBackResponse(request)
    if exc.message:
        response = response.with_error(exc.message)
    return response


async def default_http_error_handler(request: Request, exc: HTTPException) -> Response:
    if request.wants_json:
        data = {'message': exc.message, 'errors': getattr(exc, 'errors', {})}
        if request.app.debug:
            exception_type = f'{exc.__class__.__module__}.{exc.__class__.__qualname__}'
            data['exception_type'] = exception_type
            data['exception'] = repr(exc)
        return JSONResponse(data, status_code=exc.status_code)
    elif request.app.debug:
        if exc.status_code in {204, 304}:
            return Response(b"", status_code=exc.status_code)
        return PlainTextResponse(exc.detail, status_code=exc.status_code)
    else:
        content = _renderer.get_template('errors/http_error.html').render({'request': request, 'exc': exc})
        return HTMLResponse(content, status_code=exc.status_code)


_default_error_handlers = {
    BaseHTTPException: default_http_error_handler,
}


class ErrorHandlers:
    def __init__(self, handlers: dict) -> None:
        self.handlers: dict = {**_default_error_handlers, **handlers}

    def get_for_exception(self, exc: Exception) -> t.Optional[ErrorHandler]:
        class_name = type(exc)
        status_code = exc.status_code if isinstance(exc, BaseHTTPException) else 0
        if status_code in self.handlers:
            return self.handlers[status_code]

        for cls in class_name.__mro__:
            if cls in self.handlers:
                return self.handlers[cls]

        return None


class ExceptionMiddleware:
    def __init__(self, app: ASGIApp, handlers: t.Dict[t.Union[int, t.Type[Exception]], ErrorHandler]) -> None:
        self.app = app
        self.handlers = ErrorHandlers(handlers)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope['type'] != 'http':
            return await self.app(scope, receive, send)

        response_started = False

        async def sender(message: Message) -> None:
            nonlocal response_started

            if message["type"] == "http.response.start":
                response_started = True
            await send(message)

        try:
            await self.app(scope, receive, sender)
        except Exception as exc:
            handler = self.handlers.get_for_exception(exc)
            if handler is None:
                raise exc

            if response_started:
                msg = "Caught handled exception, but response already started."
                raise RuntimeError(msg) from exc

            request = Request(scope, receive=receive)
            if inspect.iscoroutinefunction(handler):
                response = await handler(request, exc)
            else:
                response = await run_in_threadpool(handler, request, exc)
            await response(scope, receive, sender)



import jinja2
import typing

from kupala.requests import Request
from kupala.responses import Response


class RenderError(Exception):
    """Base class for all renderer classes."""


class JinjaRenderer:
    def __init__(self, env: jinja2.Environment) -> None:
        self._env = env

    def render(self, template_name: str, context: typing.Mapping[str, typing.Any] = None) -> str:
        return self._env.get_template(template_name).render(context)


class TemplateResponse(Response):
    def __init__(
        self,
        request: Request,
        template_name: str,
        context: typing.Mapping[str, typing.Any] = None,
        status_code: int = 200,
        media_type: str = 'text/html',
        headers: dict = None,
    ) -> None:
        full_context = {}
        try:
            full_context.update(request.state.template_context)
        except AttributeError:
            pass

        full_context.update(context or {})
        content = request.app.renderer.render(template_name, full_context)
        super().__init__(content, status_code, headers, media_type)



from __future__ import annotations

import logging
import traceback
import typing
from contextlib import AsyncExitStack
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.middleware import MiddlewareStack
from kupala.middleware.errors import ServerErrorMiddleware
from kupala.middleware.exception import ErrorHandler, ExceptionMiddleware
from kupala.requests import Request
from kupala.responses import Response
from kupala.routing import Router, Routes

if typing.TYPE_CHECKING:  # pragma: nocover
    from kupala.application import Kupala


class ASGIHandler:
    def __init__(
        self,
        app: Kupala,
        *,
        debug: bool,
        routes: Routes,
        middleware: MiddlewareStack,
        error_handlers: dict[typing.Type[Exception] | int, typing.Optional[ErrorHandler]],
        lifespan_handlers: list[typing.Callable[[Kupala], typing.AsyncContextManager]],
        exception_handler: typing.Callable[[Request, Exception], Response] | None,
        request_class: typing.Type[Request] = None,
    ) -> None:
        self.app = app
        self.debug = debug
        self.middleware = middleware
        self.routes = Routes(routes or [])
        self.error_handlers = error_handlers
        self.exception_handler = exception_handler
        self.lifespan_handlers = lifespan_handlers
        self.request_class = request_class or Request

        middleware.use(ExceptionMiddleware, handlers=error_handlers)
        middleware.top(ServerErrorMiddleware, debug=debug, handler=self.exception_handler)
        self.asgi_app: ASGIApp = Router(routes=routes)
        for mw in reversed(middleware):
            self.asgi_app = mw.wrap(self.asgi_app)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope['type'] in {'http', 'websocket', 'lifespan'}
        if scope['type'] == 'lifespan':
            await self.lifespan_handler(scope, receive, send)
            return

        scope['app'] = self.app
        scope['state'] = {}
        if scope['type'] == 'http':
            request = self.request_class(scope, receive, send)
            scope['request'] = request
        await self.asgi_app(scope, receive, send)

    async def lifespan_handler(self, scope: Scope, receive: Receive, send: Send) -> None:
        started = False
        try:
            await receive()
            async with AsyncExitStack() as stack:
                for hook in self.lifespan_handlers:
                    await stack.enter_async_context(hook(self.app))
                await send({'type': 'lifespan.startup.complete'})
                started = True
                await receive()
        except BaseException as ex:
            logging.exception(ex)
            logging.error(f'Application failed to boot: {ex}.')
            text = traceback.format_exc()
            if started:
                await send({'type': 'lifespan.shutdown.failed', 'message': text})
            else:
                await send({'type': 'lifespan.startup.failed', 'message': text})
        else:
            await send({'type': 'lifespan.shutdown.complete'})



from __future__ import annotations

import anyio
import click
import functools
import inspect
import typing
import typing as t

if typing.TYPE_CHECKING:
    from kupala.app.base import BaseApp


def create_dispatcher(app: BaseApp, fn: t.Callable) -> t.Callable:
    fn_kwargs = t.get_type_hints(fn)

    @functools.wraps(fn)
    def dispatcher(**kwargs: t.Any) -> t.Any:
        for name, class_name in fn_kwargs.items():
            if hasattr(class_name, 'from_app'):
                kwargs[name] = class_name.from_app(app)

        if inspect.iscoroutinefunction(fn):

            async def async_dispatcher() -> t.Any:
                return await fn(**kwargs)

            return anyio.run(async_dispatcher)
        else:
            return fn(**kwargs)

    return dispatcher


def wrap_command(app: BaseApp, command: t.Union[click.Command, t.Callable]) -> click.Command:
    if isinstance(command, click.Group):
        if not command.callback:
            return command
        command.callback = create_dispatcher(app, command.callback)
        command.commands = {
            command_name: wrap_command(app, subcommand) for command_name, subcommand in command.commands.items()
        }
    elif isinstance(command, click.Command):
        if not command.callback:
            return command
        command.name = command.name.replace('-command', '') if command.name else command.name
        command.callback = create_dispatcher(app, command.callback)
    else:
        name = command.__name__.replace('_command', '')
        return click.Command(name, callback=create_dispatcher(app, command))
    return command


class ConsoleApplication:
    def __init__(self, app: BaseApp, commands: list[click.Command]) -> None:
        self.app = app
        self.commands = commands

    def run(self) -> int:
        @click.group()
        @click.pass_context
        def console_app(ctx: click.Context) -> None:
            ctx.ensure_object(dict)
            ctx.obj['app'] = self.app

        for command in self.commands:
            command = wrap_command(self.app, command)
            console_app.add_command(command)
        return console_app()



import typing as t
from starlette.exceptions import HTTPException as BaseHTTPException


class KupalaError(Exception):
    """Base class for all Kupala framework errors."""


class HTTPException(BaseHTTPException, KupalaError):
    message: t.Optional[str] = None

    def __init__(self, status_code: int, message: str = None) -> None:  # pragma: nocover
        self.message = message or self.message
        super().__init__(status_code=status_code, detail=message)  # type: ignore


class _BasePredefinedHTTPException(HTTPException):
    status_code: int

    def __init__(self, message: str = None) -> None:
        super().__init__(self.status_code, message)


class BadRequest(_BasePredefinedHTTPException):
    """The server cannot or will not process the request due to an apparent client error
    (e.g., malformed request syntax, size too large, invalid request message framing,
    or deceptive request routing)."""

    status_code = 400


class NotAuthenticated(_BasePredefinedHTTPException):
    """Similar to 403 Forbidden, but specifically for use when authentication is required
    and has failed or has not yet been provided."""

    status_code = 401


class PermissionDenied(_BasePredefinedHTTPException):
    """The request contained valid data and was understood by the server, but the server is refusing action.
    This may be due to the user not having the necessary permissions
    for a resource or needing an account of some sort, or attempting a prohibited action
    (e.g. creating a duplicate record where only one is allowed)."""

    status_code = 403


class PageNotFound(_BasePredefinedHTTPException):
    """The requested resource could not be found but may be available in the future.
    Subsequent requests by the client are permissible."""

    status_code = 404


class MethodNotAllowed(_BasePredefinedHTTPException):
    """A request method is not supported for the requested resource; for example,
    a GET request on a form that requires data to be presented via POST, or a PUT request on a read-only resource."""

    status_code = 405


class NotAcceptable(_BasePredefinedHTTPException):
    """The requested resource is capable of generating only content
    not acceptable according to the Accept headers sent in the request."""

    status_code = 406


class Conflict(_BasePredefinedHTTPException):
    """Indicates that the request could not be processed because of conflict in the current state of the resource,
    such as an edit conflict between multiple simultaneous updates."""

    status_code = 409


class UnsupportedMediaType(_BasePredefinedHTTPException):
    """The request entity has a media type which the server or resource does not support. For example,
    the client uploads an image as image/svg+xml, but the server requires that images use a different format."""

    status_code = 415


class UnprocessableEntity(_BasePredefinedHTTPException):
    """Indicates that the server is unwilling to risk processing a request that might be replayed."""

    status_code = 422


class TooManyRequests(_BasePredefinedHTTPException):
    """The user has sent too many requests in a given amount of time. Intended for use with rate-limiting schemes."""

    status_code = 429


class ValidationError(KupalaError):
    """Raised when data is not valid by any criteria."""

    def __init__(
        self,
        message: str = None,
        errors: t.Mapping[str, list[str]] = None,
    ) -> None:
        super().__init__(message)
        self.message = message
        self.errors = errors or {}



import anyio
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.responses import PlainTextResponse


class TimeoutMiddleware:
    def __init__(self, app: ASGIApp, timeout: int = 30) -> None:
        self.app = app
        self.timeout = timeout

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        try:
            async with anyio.fail_after(self.timeout):
                await self.app(scope, receive, send)
        except TimeoutError:
            response = PlainTextResponse('Gateway Timeout', 504)
            await response(scope, receive, send)



from dataclasses import dataclass

import typing as t

from kupala.middleware import Middleware


@dataclass
class ResourceAction:
    path: str
    methods: list[str]
    path_name: str
    scope: t.Literal['collection', 'object']
    include_in_schema: bool
    middleware: t.Sequence[Middleware]

    @property
    def is_for_object(self) -> bool:
        return self.scope == 'object'


def action(
    path: str,
    methods: list[str] = None,
    path_name: str = None,
    scope: t.Literal['object', 'collection'] = 'collection',
    include_in_schema: bool = True,
    middleware: t.Sequence[Middleware] = None,
) -> t.Callable:
    """Mark resource class method as custom action."""
    assert path.startswith("/"), "Routed paths must start with '/'"

    def wrapper(fn: t.Callable) -> t.Callable:
        setattr(
            fn,
            'action',
            ResourceAction(
                path=path,
                methods=methods or ['GET', 'HEAD'],
                path_name=path_name or fn.__name__,
                scope=scope,
                include_in_schema=include_in_schema,
                middleware=middleware or [],
            ),
        )
        return fn

    return wrapper


def get_resource_action(method: t.Callable) -> t.Optional[ResourceAction]:
    """Get resource action spec from resource method."""
    return getattr(method, 'action', None)



from __future__ import annotations

import typing as t

from kupala.application import Kupala
from kupala.requests import Request

_T = t.TypeVar('_T')


def to_request_injectable(klass: t.Type[_T], factory: t.Callable[[Request], _T]) -> None:
    """Convert regular class into a request injectable."""

    def augmentation(cls: t.Type[_T], request: Request) -> _T:
        return factory(request)

    setattr(klass, 'from_request', classmethod(augmentation))


def to_app_injectable(klass: t.Type[_T], factory: t.Callable[[Kupala], _T]) -> None:
    """Convert regular class into app injectable."""

    def augmentation(cls: t.Type[_T], app: Kupala) -> _T:
        return factory(app)

    setattr(klass, 'from_app', classmethod(augmentation))



import hashlib
import hmac
import os
import typing as t
from itsdangerous import BadData, SignatureExpired, URLSafeTimedSerializer
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.exceptions import PermissionDenied
from kupala.requests import Request

CSRF_SESSION_KEY = '_csrf_token'
CSRF_HEADER = 'x-csrf-token'
CSRF_QUERY_PARAM = 'csrf-token'
CSRF_POST_FIELD = '_token'


class CSRFError(Exception):
    pass


class TokenMissingError(CSRFError):
    pass


class TokenExpiredError(CSRFError):
    pass


class TokenDecodeError(CSRFError):
    pass


class TokenMismatchError(CSRFError):
    pass


def get_generate_random(length: int = 64) -> str:
    return hashlib.sha1(os.urandom(length)).hexdigest()


def generate_token(secret_key: str, data: str) -> str:
    return hmac.new(secret_key.encode(), data.encode(), 'sha256').hexdigest()


def validate_csrf_token(
    session_token: str,
    timed_token: str,
    secret_key: str,
    salt: str = '_csrf_',
    max_age: int = None,
) -> bool:
    if not timed_token or not session_token:
        raise TokenMissingError('CSRF token is missing.')

    try:
        serializer = URLSafeTimedSerializer(secret_key, salt=salt)
        raw_token = serializer.loads(timed_token, max_age=max_age)
    except SignatureExpired:
        raise TokenExpiredError('CSRF token has expired.')
    except BadData:
        raise TokenDecodeError('CSRF token is invalid.')

    if not hmac.compare_digest(session_token, raw_token):
        raise TokenMismatchError('CSRF tokens do not match.')
    return True


class CSRFMiddleware:
    exclude_urls: t.Optional[t.Iterable[str]] = None
    safe_methods: t.Iterable[str] = ['get', 'head', 'options']

    def __init__(
        self,
        app: ASGIApp,
        secret_key: str,
        salt: str = '_csrf_',
        exclude_urls: t.Iterable[str] = None,
        max_age: int = 3600,
    ):
        self.app = app
        self._exclude_urls = exclude_urls or self.exclude_urls or []
        self._secret_key = str(secret_key)
        self._salt = salt
        self._max_age = max_age

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if 'session' not in scope:
            raise CSRFError('CsrfMiddleware requires SessionMiddleware.')

        await scope['session'].load()
        request = scope['request']
        if CSRF_SESSION_KEY not in request.session:
            token = get_generate_random()
            csrf_token = generate_token(self._secret_key, token)
            request.session[CSRF_SESSION_KEY] = csrf_token

        serializer = URLSafeTimedSerializer(self._secret_key, self._salt)
        timed_token = str(serializer.dumps(request.session[CSRF_SESSION_KEY], self._salt))
        request.state.csrf_token = request.session[CSRF_SESSION_KEY]
        request.state.csrf_timed_token = timed_token

        try:
            request.state.template_context.update(
                {
                    'csrf_token': get_csrf_token(request),
                    'csrf_input': get_csrf_input(request),
                }
            )
        except AttributeError:
            # TemplateContextMiddleware is not installed
            pass

        if self.should_check_token(request):
            try:
                validate_csrf_token(
                    session_token=request.session[CSRF_SESSION_KEY],
                    timed_token=await self.get_csrf_token(request),
                    secret_key=self._secret_key,
                    salt=self._salt,
                    max_age=self._max_age,
                )
            except CSRFError:
                raise PermissionDenied('CSRF token is invalid.')
        await self.app(scope, receive, send)

    async def get_csrf_token(self, request: Request) -> str:
        from_headers = request.headers.get(CSRF_HEADER)
        from_query = request.query_params.get(CSRF_QUERY_PARAM)
        from_form_data = None
        if request.method.lower() in ['post', 'put', 'patch', 'delete']:
            form_data = await request.form()
            from_form_data = form_data.get(CSRF_POST_FIELD)
        return from_query or from_form_data or from_headers

    def should_check_token(self, request: Request) -> bool:
        return not any(
            [
                request.method.lower() in self.safe_methods,
                request.url_matches(*self._exclude_urls),
                request.full_url_matches(*self._exclude_urls),
            ]
        )


def get_csrf_token(request: Request) -> t.Optional[str]:
    return request.state.csrf_timed_token if hasattr(request.state, 'csrf_timed_token') else None


def get_csrf_input(request: Request) -> str:
    token = get_csrf_token(request) or ''
    return f'<input type="hidden" name="{CSRF_POST_FIELD}" value="{token}">'



from dataclasses import dataclass, field

import inspect
import typing
import typing as t
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from starlette.concurrency import run_in_threadpool
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.middleware import Middleware
from kupala.requests import Request
from kupala.responses import EmptyResponse, HTMLResponse, JSONResponse, PlainTextResponse, Response
from kupala.templating import TemplateResponse
from kupala.utils import run_async


def _callable_name(injection: typing.Any) -> str:
    class_name = injection.__name__ if inspect.isclass(injection) else getattr(injection, '__name__', repr(injection))
    module_name = getattr(injection, '__module__', '')
    return f'{module_name}.{class_name}{inspect.signature(injection)}'


@dataclass
class ActionConfig:
    """Keeps endpoint configuration."""

    methods: t.List[str] = field(default_factory=list)
    template: t.Optional[str] = None
    middleware: t.Optional[t.Sequence[Middleware]] = None


def action_config(
    methods: t.List[str] = None,
    template: str = '',
    middleware: t.Sequence[Middleware] = None,
) -> t.Callable:
    """Use this decorator to configure endpoint parameters."""
    allowed_methods = methods or ['GET', 'HEAD']

    def wrapper(fn: t.Callable) -> t.Callable:
        setattr(
            fn, '__action_config__', ActionConfig(methods=allowed_methods, template=template, middleware=middleware)
        )
        return fn

    return wrapper


def get_action_config(endpoint: t.Callable) -> t.Optional[ActionConfig]:
    """Return endpoint config if defined otherwise return None."""
    return getattr(endpoint, '__action_config__', None)


def detect_request_class(endpoint: t.Callable) -> t.Type[Request]:
    """Detect which request class to use for this endpoint.
    If endpoint does not have `request` argument, or it is not type-hinted
    then default request class returned."""
    args = t.get_type_hints(endpoint)
    return args.get('request', Request)


async def resolve_injections(
    request: Request,
    endpoint: t.Callable,
    sync_stack: ExitStack,
    async_stack: AsyncExitStack,
) -> dict[str, t.Any]:
    """Read endpoint signature and extract injections types.
    These injections will be resolved into actual service instances.
    Dependency injections and path parameters are merged.

    Return value of `from_request` can be a generator. In this case we convert it into context manager
    and add to sync/async exit stack.
    """
    injections = {}

    args = t.get_type_hints(endpoint)
    for arg_name, arg_type in args.items():
        if arg_name == 'return':
            continue

        if arg_type == type(request):
            injections[arg_name] = request
            continue

        if arg_name in request.path_params:
            injections[arg_name] = request.path_params[arg_name]
            continue

        if callback := getattr(arg_type, 'from_request', None):
            if inspect.isgeneratorfunction(callback):
                injections[arg_name] = sync_stack.enter_context(contextmanager(callback)(request))
                continue

            if inspect.isasyncgenfunction(callback):
                injections[arg_name] = await async_stack.enter_async_context(asynccontextmanager(callback)(request))
                continue

            injections[arg_name] = await run_async(callback, request=request)
            continue

        if callback := getattr(arg_type, 'from_app', None):
            injections[arg_name] = await run_async(callback, app=request.app)
            continue

        raise TypeError(f'Injection "{arg_name}" cannot be processed in {_callable_name(endpoint)}.')
    return injections


def _guess_response_type(
    request: Request,
    action_config: ActionConfig | None,
    content: t.Any,
    status: int,
    headers: dict[str, str],
) -> Response:
    """When endpoint returns non-response instances, then we try to do our best to guess what response type
    we should send to the client."""
    if request.wants_json:
        return JSONResponse(content=content, status_code=status, headers=headers)

    if 'text/html' in request.headers['accept']:
        if action_config and action_config.template:
            return TemplateResponse(
                request,
                template_name=action_config.template,
                context=content,
                status_code=status,
                headers=headers,
            )
        return HTMLResponse(content, status_code=status, headers=headers)

    # default is text/plain, but we can serialize only several types
    return PlainTextResponse(str(content), status_code=status, headers=headers)


def handle_endpoint_result(request: Request, result: t.Any, action_config: ActionConfig | None) -> ASGIApp:
    """Takes return value of endpoint execution and converts it into response instance."""
    if result is None:
        return EmptyResponse()

    if isinstance(result, Response):
        return result

    if inspect.isfunction(result):  # may be request_response function aka ASGI callable
        return result

    status = 200
    headers = {}
    content = result

    if isinstance(result, tuple):
        if len(result) == 3:
            content, status, headers = result
        elif len(result) == 2:
            content, status = result

    return _guess_response_type(request, action_config, content, status, headers)


async def dispatch_endpoint(scope: Scope, receive: Receive, send: Send, endpoint: t.Callable) -> ASGIApp:
    """Call endpoint callable resolving all dependencies. Will return response."""
    request_class = detect_request_class(endpoint)
    request = request_class(scope, receive, send)
    action_config = get_action_config(endpoint)

    with ExitStack() as sync_stack:
        async with AsyncExitStack() as async_stack:
            args = await resolve_injections(request, endpoint, sync_stack, async_stack)
            if inspect.iscoroutinefunction(endpoint):
                response = await endpoint(**args)
            else:
                response = await run_in_threadpool(endpoint, **args)
    return handle_endpoint_result(request, response, action_config)



from __future__ import annotations

import click
import contextvars as cv
import jinja2
import logging
import secrets
import typing
from starlette.datastructures import State
from starlette.routing import BaseRoute
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.asgi import ASGIHandler
from kupala.config import Config
from kupala.console.application import ConsoleApplication
from kupala.contracts import TemplateRenderer
from kupala.dotenv import DotEnv
from kupala.extensions import Authentication, Mail, Passwords, Renderer, Storages
from kupala.middleware import Middleware, MiddlewareStack
from kupala.middleware.exception import ErrorHandler
from kupala.requests import Request
from kupala.responses import Response
from kupala.routing import Routes
from kupala.storages.storages import Storage
from kupala.templating import JinjaRenderer
from kupala.utils import resolve_path


class Kupala:
    request_class = Request

    def __init__(
        self,
        *,
        secret_key: str = None,
        debug: bool = None,
        env_file: str = '.env',
        middleware: list[Middleware] = None,
        routes: list[BaseRoute] | Routes | None = None,
        request_class: typing.Type[Request] | None = None,
        error_handlers: dict[typing.Type[Exception] | int, typing.Optional[ErrorHandler]] = None,
        lifespan_handlers: list[typing.Callable[[Kupala], typing.AsyncContextManager[None]]] = None,
        exception_handler: typing.Callable[[Request, Exception], Response] | None = None,
        template_dir: str | list[str] = 'templates',
        commands: list[click.Command] = None,
        storages: dict[str, Storage] = None,
        renderer: TemplateRenderer = None,
    ) -> None:
        # base configuration
        self.secret_key = secret_key
        self.request_class = request_class or self.request_class or Request

        # read environment and set default variables
        self.environ = DotEnv([env_file])
        self.secret_key = secret_key or self.environ.get('SECRET_KEY', default=None, check_file=True)
        if self.secret_key is None:
            self.secret_key = secrets.token_hex(128)
            logging.warning(
                'Secret key is not defined. Will generate temporary secret. '
                'This value will be reset after application restart.'
            )

        self.debug = debug if debug is not None else self.environ.bool('APP_DEBUG', False)
        self.environment = self.environ.get('APP_ENV', 'production')

        # ASGI related
        self.error_handlers = error_handlers or {}
        self.lifespan = lifespan_handlers or []
        self.exception_handler = exception_handler
        self.middleware = MiddlewareStack(middleware)
        self.routes = Routes(routes or [])

        self.template_dirs = [template_dir] if isinstance(template_dir, str) else template_dir or []

        # assign core components
        self.config = Config()
        self.state = State()
        self.passwords = Passwords()
        self.mail = Mail()
        self.auth = Authentication(self)
        self.storages = Storages(storages)
        self.commands = commands or []

        self.jinja_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(
                searchpath=[resolve_path(template_path) for template_path in self.template_dirs]
            )
        )
        self.jinja_env.policies["json.dumps_kwargs"] = {"ensure_ascii": False, "sort_keys": True}
        self.renderer = Renderer(renderer or JinjaRenderer(self.jinja_env))

        set_current_application(self)
        self.bootstrap()

        # ASGI app instance
        self._asgi_app: ASGIApp | None = None

    def bootstrap(self) -> None:
        pass

    def apply_middleware(self) -> None:
        pass

    def create_asgi_app(self) -> ASGIApp:
        self.apply_middleware()
        return ASGIHandler(
            app=self,
            debug=self.debug,
            routes=self.routes,
            middleware=self.middleware,
            error_handlers=self.error_handlers,
            lifespan_handlers=self.lifespan,
            exception_handler=self.exception_handler,
            request_class=self.request_class,
        )

    def cli(self) -> int:
        set_current_application(self)
        app = ConsoleApplication(self, self.commands)
        return app.run()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if self._asgi_app is None:
            self._asgi_app = self.create_asgi_app()
        await self._asgi_app(scope, receive, send)


_current_app: cv.ContextVar[Kupala] = cv.ContextVar('_current_app')


def set_current_application(app: Kupala) -> None:
    _current_app.set(app)


def get_current_application() -> Kupala:
    return _current_app.get()



import os
from starlette.types import Receive, Scope, Send

from kupala.responses import FileResponse, PlainTextResponse, RedirectResponse, Response
from kupala.storages.storages import Storage


class FileServer:
    def __init__(self, storage: str, as_attachment: bool = True) -> None:
        self.storage = storage
        self.as_attachment = as_attachment

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "http"
        path = self.get_path(scope)
        response = await self.get_response(path, scope)
        await response(scope, receive, send)

    async def get_response(self, path: str, scope: Scope) -> Response:
        if scope["method"] not in ("GET", "HEAD"):
            return PlainTextResponse("Method Not Allowed", status_code=405)

        url = await self.get_disk(scope).url(path)
        if url.startswith('http'):
            return RedirectResponse(url, 307)
        path = self.get_disk(scope).abspath(path)
        return FileResponse(path, inline=not self.as_attachment)

    def get_path(self, scope: Scope) -> str:
        return os.path.normpath(os.path.join(*scope["path"].split("/")))

    def get_disk(self, scope: Scope) -> Storage:
        return scope['app'].storages.get(self.storage)



import typing as t

from kupala.application import get_current_application
from kupala.requests import Request
from kupala.response_factories import ResponseFactory


def response(request: Request, status_code: int = 200, headers: dict = None) -> ResponseFactory:
    return ResponseFactory(request, status_code, headers)


def render_to_string(template_name: str, context: t.Dict = None) -> str:
    app = get_current_application()
    return app.renderer.render(template_name, context)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kupala', 'kupala.console', 'kupala.middleware', 'kupala.storages']

package_data = \
{'': ['*'], 'kupala': ['templates/errors/*']}

install_requires = \
['Babel>=2.9.1,<3.0.0',
 'Jinja2>=3.0,<4.0',
 'click>=8.0,<9.0',
 'deesk>=0.1,<0.2',
 'imia>=0.3,<0.4',
 'mailers>=2.0.0,<3.0.0',
 'passlib>=1.7.4,<2.0.0',
 'python-dotenv>=0.19,<0.20',
 'python-multipart>=0.0.5,<0.0.6',
 'starlette>=0.17,<0.18',
 'starsessions>=1.2,<2.0',
 'toronado>=0.1.0,<0.2.0']

extras_require = \
{'s3': ['aioboto3>=9.2,<10.0']}

setup_kwargs = {
    'name': 'kupala',
    'version': '0.12.0',
    'description': 'A modern web framework for Python.',
    'long_description': '# Kupala Framework\n\nA modern web framework for Python.\n\n![PyPI](https://img.shields.io/pypi/v/kupala)\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/alex-oleshkevich/kupala/Lint)\n![GitHub](https://img.shields.io/github/license/alex-oleshkevich/kupala)\n![Libraries.io dependency status for latest release](https://img.shields.io/librariesio/release/pypi/kupala)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/kupala)\n![GitHub Release Date](https://img.shields.io/github/release-date/alex-oleshkevich/kupala)\n![Lines of code](https://img.shields.io/tokei/lines/github/alex-oleshkevich/kupala)\n\n## Installation\n\nInstall `kupala` using PIP or poetry:\n\n```bash\npip install kupala\n# or\npoetry add kupala\n```\n\n## Features\n\n-   TODO\n\n## Quick start\n\nSee example application in `examples/` directory of this repository.\n',
    'author': 'Alex Oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alex-oleshkevich/kupala',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import typing
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.contracts import ContextProcessor
from kupala.requests import Request


class TemplateContextMiddleware:
    """For every request this middleware will call given context processors.
    There return value will be merged and stored in `request.state.template_context` attribute.
    TemplateResponse will merge it with its own context providing a final template context."""

    def __init__(self, app: ASGIApp, context_processors: list[ContextProcessor] = None) -> None:
        self.app = app
        self.context_processors = context_processors or []

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope['type'] != 'http':  # pragma: nocover
            await self.app(scope, receive, send)
            return

        request = Request(scope, receive, send)
        context: dict[str, typing.Any] = {
            'request': request,
            'config': request.app.config,
            'url': request.url,
            'app': request.app,
            'form_errors': request.form_errors,
            'old_input': request.old_input,
        }
        request = Request(scope, receive, send)
        for processor in self.context_processors:
            context.update(processor(request))
        scope['state'].setdefault('template_context', {})
        scope['state']['template_context'].update(context)
        await self.app(scope, receive, send)



from __future__ import annotations

import mimetypes
import os
import pathlib
import re
import typing
import typing as t
import uuid
from babel.core import Locale
from imia import UserLike, UserToken
from starlette import datastructures as ds, requests
from starlette.datastructures import State
from starlette.requests import empty_receive, empty_send
from starlette.types import Receive, Scope, Send
from starsessions import Session

from kupala.storages.storages import Storage

if t.TYPE_CHECKING:
    from .app.base import BaseApp


class OldFormInput(t.Mapping):
    def __init__(self, data: dict) -> None:
        self._data = data

    def get(self, key: str, default: t.Any = None) -> t.Any:
        return self._data.get(key, default)

    def to_json(self) -> dict:
        return self._data

    def __getitem__(self, item: str) -> t.Any:
        return self._data[item]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> t.Iterator[t.Tuple[str, t.Any]]:
        return iter(self._data)

    def __contains__(self, item: object) -> bool:
        return item in self._data

    def __repr__(self) -> str:
        return f'<OldFormInput: data={self._data}>'

    __call__ = get


class FormErrors(t.Mapping):
    def __init__(self, errors: t.Mapping[str, list[str]]) -> None:
        self.field_errors = errors

    def get(self, key: str, default: t.Any = None) -> t.Optional[list[str]]:
        return self.field_errors.get(key, default)

    def to_json(self) -> dict:
        return {
            'field_errors': self.field_errors,
        }

    def keys(self) -> t.KeysView[str]:
        return self.field_errors.keys()

    def __bool__(self) -> bool:
        return bool(self.field_errors)

    def __getitem__(self, item: str) -> list[str]:
        return self.field_errors[item]

    def __contains__(self, item: object) -> bool:
        return item in self.field_errors

    def __iter__(self) -> t.Iterator:
        return iter(self.field_errors)

    def __len__(self) -> int:
        return len(self.field_errors)

    def __repr__(self) -> str:
        return f'<FormErrors: field_errors={self.field_errors}>'

    __call__ = get


class QueryParams(requests.QueryParams):
    def get_bool(self, key: str, default: bool = None) -> t.Optional[bool]:
        value = self.get(key, None)
        return value.lower() in ['t', 'true', 'yes', 'on', '1'] if value is not None else default

    def get_list(self, key: str, subcast: t.Callable = None) -> list[str]:
        items = self.getlist(key)
        if subcast:
            return list(map(subcast, items))
        return items

    def get_int(self, key: str, default: int = None) -> t.Optional[int]:
        value: str = self.get(key, None)
        return int(value) if value is not None and value.isnumeric() else default


class FormData(requests.FormData):
    pass


class UploadFile(ds.UploadFile):
    def __init__(self, request: Request, filename: str, file: typing.IO = None, content_type: str = "") -> None:
        self.request = request
        super().__init__(filename, file, content_type)

    @classmethod
    def from_base_upload_file(cls, upload_file: ds.UploadFile, request: Request) -> UploadFile:
        return UploadFile(
            filename=upload_file.filename,
            file=upload_file.file,
            content_type=upload_file.content_type,
            request=request,
        )

    async def save(self, directory: t.Union[str, os.PathLike], filename: str = None, disk: str = None) -> str:
        uploaded_filename = pathlib.Path(self.filename)
        extension = pathlib.Path(uploaded_filename).suffix
        base_name = os.path.basename(uploaded_filename).replace(extension, '')
        prefix = str(uuid.uuid4().fields[-1])[:5]
        if not extension:
            extension = '.' + (mimetypes.guess_extension(self.content_type) or 'bin')
        extension = extension.lstrip('.')
        suggested_filename = f'{prefix}_{base_name}.{extension}'

        file_path = os.path.join(directory, filename or suggested_filename)

        storage: Storage
        if disk:
            storage = self.request.app.storages.get(disk)
        else:
            storage = self.request.app.storages.get_default()
        await storage.put(file_path, await self.read())
        return file_path

    async def read(self, size: int = -1) -> bytes:
        return t.cast(bytes, await super().read(size))

    async def read_string(self, size: int = -1, encoding: str = 'utf8', errors: str = 'strict') -> str:
        return (await self.read(size)).decode(encoding, errors)


class FilesData:
    def __init__(self, files: FormData) -> None:
        self._files = files

    def get(self, key: str, default: t.Any = None) -> t.Optional[UploadFile]:
        return self._files.get(key, default)

    def getlist(self, key: str) -> list[UploadFile]:
        return t.cast(list[UploadFile], self._files.getlist(key))


class Cookies(dict):
    pass


class Headers(requests.Headers):
    pass


_A = t.TypeVar('_A', bound='BaseApp')
_S = t.TypeVar('_S', bound=State)


class Request(requests.Request, t.Generic[_A, _S]):
    def __new__(cls, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send) -> Request:
        if 'request' not in scope:
            instance = super().__new__(cls)
            instance.__init__(scope, receive, send)  # type: ignore
            scope['request'] = instance
        elif scope['request'].__class__ != cls:
            # view function uses custom request class
            request = scope['request']
            request.__class__ = cls
            scope['request'] = request
        return scope['request']

    @property
    def query_params(self) -> QueryParams:
        return QueryParams(super().query_params)

    @property
    def app(self) -> _A:
        return self.scope['app']

    @property
    def auth(self) -> UserToken:
        assert "auth" in self.scope, "AuthenticationMiddleware must be installed to access request.auth"
        return self.scope["auth"]

    @property
    def user(self) -> UserLike:
        return self.auth.user

    @property
    def wants_json(self) -> bool:
        """Test if request sends Accept header
        with 'application/json' value."""
        return "application/json" in self.headers.get("accept", "")

    @property
    def is_json(self) -> bool:
        return "application/json" in self.headers.get("content-type", "")

    @property
    def is_xhr(self) -> bool:
        """
        Is true when the request is a XMLHttpRequest.
        It works if JavaScript's HTTP client sets an X-Requested-With HTTP header.
        Known frameworks:
        http://en.wikipedia.org/wiki/List_of_Ajax_frameworks#JavaScript
        :return:
        """
        return self.headers.get("x-requested-with", None) == "XMLHttpRequest"

    @property
    def ip(self) -> str:
        """Returns the IP address of user."""
        return self.client.host

    @property
    def secure(self) -> bool:
        """Determine if the request served over SSL."""
        return self.scope["scheme"] == "https"

    @property
    def is_post(self) -> bool:
        """Test if request was made using POST command."""
        return self.method.upper() == "POST"

    @property
    def session(self) -> Session:
        assert "session" in self.scope, "SessionMiddleware must be installed to access request.session"
        return self.scope['session']

    @property
    def locale(self) -> Locale:
        assert "locale" in self.scope, "LocaleMiddleware must be installed to access request.locale"
        return self.scope['locale']

    @property
    def language(self) -> str:
        return self.locale.language

    @property
    def cookies(self) -> Cookies:
        return Cookies(**super().cookies)

    @property
    def headers(self) -> Headers:
        return Headers(super().headers)

    @property
    def old_input(self) -> OldFormInput:
        """Get previous form input. This value does not include uploaded files."""
        if '_form_old_input' not in self.scope:
            old_form_input = OldFormInput({})
            if 'session' in self.scope:
                old_form_input = OldFormInput(self.session.pop('_form_old_input', {}))
            self.scope['_form_old_input'] = old_form_input
        return self.scope['_form_old_input']

    @property
    def state(self) -> _S:
        return t.cast(_S, super().state)

    async def remember_form_data(self) -> None:
        """Flush current form data into session so it can be used on the next page to render form errors."""
        if 'session' in self.scope:
            data = await self.form()
            self.session['_form_old_input'] = {k: v for k, v in data.items() if not isinstance(v, ds.UploadFile)}

    @property
    def form_errors(self) -> FormErrors:
        """Get form errors generated on the previous page."""
        if '_form_error' not in self.scope:
            form_error = FormErrors({})
            if 'session' in self.scope:
                form_error = FormErrors(self.session.pop('_form_field_errors', {}))
            self.scope['_form_error'] = form_error
        return self.scope['_form_error']

    def set_form_errors(self, field_errors: dict = None) -> None:
        """Flush form error message and form field errors into session.
        This data can be later retrieved via `request.form_errors` attribute."""
        if 'session' in self.scope:
            self.session['_form_field_errors'] = field_errors or {}

    def url_matches(self, *patterns: t.Union[str, t.Pattern]) -> bool:
        for pattern in patterns:
            if pattern == self.url.path:
                return True
            if re.search(pattern, self.url.path):
                return True
        return False

    def full_url_matches(self, *patterns: t.Union[str, t.Pattern]) -> bool:
        for pattern in patterns:
            if pattern == str(self.url):
                return True
            if re.match(pattern, str(self.url)):
                return True
        return False

    async def form(self) -> FormData:
        data = await super().form()
        return FormData(
            [
                (k, UploadFile.from_base_upload_file(v, self) if isinstance(v, ds.UploadFile) else v)
                for k, v in data.multi_items()
            ]
        )

    async def files(self) -> FilesData:
        data = await self.form()
        return FilesData(FormData([(k, v) for k, v in data.multi_items() if isinstance(v, UploadFile)]))

    async def data(self) -> t.Mapping:
        """Returns a request data.
        Automatically decodes JSON if Content-Type is application/json."""
        if self.is_json:
            return await self.json()
        return await self.form()

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.method} {self.url}>'
